
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "data.h"
#include "task.h"
#include "subtask.h"
#include "task_operators.h"
#include "subtask_operators.h"
#include "remove_comments.h"

#define FILE_TABLE_SIZE 100

typedef struct finfo {
                      /* file_info data type */

                      char *file_name;

                      int holder;

                      int used_as_input;

                      int entry_valid;
                     
                     } file_info;

static file_info file_table[FILE_TABLE_SIZE];



int search_available_files(char *search_for_this)
{

   int loopcnt;


   for (loopcnt = 0;
        loopcnt < FILE_TABLE_SIZE;
        loopcnt++)
     {

      if (!(file_table[loopcnt].entry_valid)) return 0; 

      if (!strcmp(search_for_this, file_table[loopcnt].file_name))
	  {
           file_table[loopcnt].used_as_input = 1;

           return (file_table[loopcnt].holder);
	  }
     }

   return 0;
}



void add_available_file(char *add_this_file, int on_this_machine)
{

   int loopcnt;


   for (loopcnt = 0;
        loopcnt < FILE_TABLE_SIZE;
        loopcnt++)
     {

      if (!(file_table[loopcnt].entry_valid)) break; 

      if (!strcmp(add_this_file, file_table[loopcnt].file_name))
	  {
           file_table[loopcnt].used_as_input = 0;

           file_table[loopcnt].holder = on_this_machine;

           return;
	  }
     }

   if (loopcnt == FILE_TABLE_SIZE)  { printf("Too many files used\n");

                                      exit(4);
                                    }

   file_table[loopcnt].entry_valid = 1;

   file_table[loopcnt].file_name = (char*) malloc((strlen(add_this_file) + 1));

   strcpy(file_table[loopcnt].file_name, add_this_file);

   file_table[loopcnt].holder = on_this_machine;
}
   

void cleanup_files(int winning_machine, subtask *completed_subtask)
{

   int loopcnt;

   int machine_list[MAX_ARCHITECTURES+1];


   for (loopcnt = 0; loopcnt < MAX_ARCHITECTURES; loopcnt++)
     {
      if ((*completed_subtask).executables[loopcnt].machine_type)
          machine_list[loopcnt] = 
             (*completed_subtask).executables[loopcnt].machine_type;

      else break;
     }

   machine_list[loopcnt] = 0;

   for (loopcnt = 0; ; loopcnt++)
      {
       if ((*completed_subtask).output[loopcnt] == NULL) return;

       VCE_cleanup((*completed_subtask).output[loopcnt], 
                   machine_list, winning_machine);

       add_available_file((*completed_subtask).output[loopcnt],
                          winning_machine);
      }
}

void copy_input_files(subtask *starting_subtask)
{

   int loopcnt;

   int machine_list[MAX_ARCHITECTURES+1];


   for (loopcnt = 0; loopcnt < MAX_ARCHITECTURES; loopcnt++)
     {
      if ((*starting_subtask).executables[loopcnt].machine_type)
          machine_list[loopcnt] = 
             (*starting_subtask).executables[loopcnt].machine_type;

      else break; 
     }

   machine_list[loopcnt] = 0;

   for (loopcnt = 0; ; loopcnt++)
     {
      if ((*starting_subtask).input[loopcnt] == NULL) break;

      VCE_copy_inputfile((*starting_subtask).input[loopcnt], machine_list,
                  search_available_files((*starting_subtask).input[loopcnt]));
     }
}

FILE *preprocess(FILE *input)
{
   FILE *new_file;


   new_file = fopen("vcerun.temp","w");

   if (new_file == NULL) {  printf("Unable to open vcerun.temp\n");

                            exit(4);
			 }

   remove_comments(input, new_file);

   fclose(input);

   fclose(new_file);

   new_file = fopen("vcerun.temp","r");

   if (new_file == NULL) {  printf("Unable to open vcerun.temp\n");

                            exit(4);
			 }

   return new_file;
}



void parse_application(FILE *input, 
                       subtask **subtask_anchor, 
                       task **task_anchor,
                       char *application)
{

   static char buffer[1000];

   char *next_operand;

   subtask *subtask_ptr;


   fscanf(input, "%s", buffer);

   if (feof(input)) 
     {
      printf("Premature end of file in application description\n");

      exit(4);
     }

   if (strlen(buffer) > 64)  { printf("Application name too long, ");

                               printf("maximum 64 characters\n");

                               exit(4);
			     }

   strcpy(application, buffer);

   fscanf(input, "%s", buffer);

   if (feof(input)) 
     {
      printf("Premature end of file in application description\n");

      exit(4);
     }

    next_operand = buffer;

    while(!feof(input)) 

      switch(parse_task_keywords(next_operand))
	  {
	  case SUBTASK: next_operand = read_subtask(input, subtask_anchor,
                                                     task_anchor);

                         break;

	  default: printf("Syntax error, unable to parse keyword %s\n",
                          next_operand);

                   exit(4);
	  }
}


void kill_slow_machines(int machine_type, subtask* completed_subtask)
{

   int loopcnt;

   void start_successors(subtask*);


   printf("%s completed on machine type %d\n", (*completed_subtask).name,
          machine_type);


   if ((*completed_subtask).status != RUNNING) return;

   for (loopcnt = 0;
        (*completed_subtask).executables[loopcnt].machine_type;
        loopcnt++)

     if ((*completed_subtask).executables[loopcnt].machine_type != 
                                                             machine_type)

    VCE_kill_executable((*completed_subtask).executables[loopcnt].machine_type,
                         (void*) completed_subtask);

   (*completed_subtask).status = COMPLETE;

   cleanup_files(machine_type, completed_subtask);

   start_successors(completed_subtask);
}  


void start_subtask(subtask *start_this)
{

   int loopcnt;

   void start_coreqs(subtask*);


   copy_input_files(start_this);

   (*start_this).status = RUNNING;
  
   for (loopcnt = 0;
        (*start_this).executables[loopcnt].machine_type != 0;
        loopcnt++)
     {

      printf("Launching %d\n", start_this);

      VCE_launch((*start_this).executables[loopcnt].machine_type, 
                 (*start_this).executables[loopcnt].file_name,
                 (*start_this).command,  
                 (*start_this).executables[loopcnt].number_of_nodes,
                 start_this);

      printf("Launched %s %s on %d nodes of machine type %d\n",
             (*start_this).executables[loopcnt].file_name,
             (*start_this).command,
             (*start_this).executables[loopcnt].number_of_nodes,
             (*start_this).executables[loopcnt].machine_type);
     }
 
   start_coreqs(start_this);

}


void start_coreqs(subtask *starting_subtask)
{

   int loopcnt;

   subtask *coreq_ptr;


   for (loopcnt = 0;
        (*starting_subtask).begin_concurrently[loopcnt] != NULL;
        loopcnt++)
     {
      coreq_ptr = (*starting_subtask).begin_concurrently[loopcnt];

      (*coreq_ptr).predecessor_count--;

      if (!((*coreq_ptr).predecessor_count)) start_subtask(coreq_ptr);
     }
}


void start_successors(subtask *completed_subtask)
{

   int loopcnt;

   subtask *successor_ptr;


   for (loopcnt = 0;
        (*completed_subtask).begin_consecutively[loopcnt] != NULL;
        loopcnt++)
     {
      successor_ptr = (*completed_subtask).begin_consecutively[loopcnt];

      (*successor_ptr).predecessor_count--;

      printf("%s complete, %s predecessor count = %d\n",
             (*completed_subtask).name,
             (*successor_ptr).name,
             (*successor_ptr).predecessor_count);


      if (!((*successor_ptr).predecessor_count)) 
        {
         printf("Starting %s\n", (*successor_ptr).name);

         start_subtask(successor_ptr);
        }
     }
}



void dispatch_application(subtask *subtask_queue)
{

   subtask *subtask_ptr = subtask_queue;

   int all_subtasks_running, done_machine_type, result;

   char *done_executable;

   subtask *completed_subtask; 


   decode_architectures(subtask_queue);


   for (; subtask_ptr != NULL; subtask_ptr = (*subtask_ptr).next_subtask)

         print_subtask(subtask_ptr);


   do {
       all_subtasks_running = 1;

       for (subtask_ptr = subtask_queue;
            subtask_ptr != NULL;
            subtask_ptr = (*subtask_ptr).next_subtask)
         {
          if (((*subtask_ptr).status != COMPLETE)
                  && ((*subtask_ptr).status != RUNNING)  
                  && ((*subtask_ptr).predecessor_count == 0))
                         { 
                           start_subtask(subtask_ptr);

                           start_coreqs(subtask_ptr); 
                         }                            

          else if ((*subtask_ptr).predecessor_count != 0)

                          all_subtasks_running = 0;
         }



     do {

       printf("About to wait\n");

       result = VCE_wait_for_status(&done_machine_type, &completed_subtask);

       printf("Done waiting, completed_subtask = %d\n", completed_subtask);

       if (result == SUCCESS) 
         {
          kill_slow_machines(done_machine_type, completed_subtask);
	 }

       else  { printf("Problem with %s on machine type %d, status = %d\n", 
                       (*completed_subtask).name,
                       done_machine_type, result); 

               exit(4);
             }

     }
     while(result != SUCCESS);

      }

   while(!all_subtasks_running);
}


int some_subtasks_are_still_running(subtask *subtask_queue)
{

   subtask *queue_ptr;


   for (queue_ptr = subtask_queue;
        queue_ptr != NULL;
        queue_ptr = (*queue_ptr).next_subtask)

      if ((*queue_ptr).status == RUNNING) break;

   if (queue_ptr == NULL) return 0;

   return 1;
}

   
void main(int argc, char *argv[])
{

   int done_machine_type, result;

   char *done_executable;

   FILE *input;

   char buffer[1000];

   char *next_operand;

   char application_name[64];

   task *task_queue_anchor = NULL;

   task *task_queue;

   subtask *subtask_queue_anchor = NULL;

   subtask *completed_subtask;


   VCE_init(argc, argv);

   input = fopen(argv[1], "r");

   if (input == NULL)  { printf("Application file did not open\n");

                        exit(4);
                      }

   /*input = preprocess(input);*/ 

   fscanf(input, "%s", buffer);

   if (feof(input))  { printf("Application statement missing\n");

                       exit(4);
                     }

   next_operand = buffer;
      
   while(!feof(input))
     
      switch(parse_task_keywords(next_operand))
	{
	case TASK: if (task_queue_anchor == NULL) 

                     task_queue_anchor = 
                               task_queue = (task*) calloc(1, sizeof(task));
                 
                   else  { (*task_queue).next_task =
                                        (task*) calloc(1, sizeof(task));

                           task_queue = (*task_queue).next_task;
                         }

                   next_operand = read_task(input, task_queue);

                   break;

	case APPLICATION: parse_application(input, 
                                            &subtask_queue_anchor,
                                            &task_queue_anchor, 
                                            application_name);

                          break;

        default: printf("Sytax error in application description, ");

                 printf("unable to parse keyword: %s\n", next_operand);

                 exit(4);
	}

   dispatch_application(subtask_queue_anchor);

   while(some_subtasks_are_still_running(subtask_queue_anchor))
     {

      printf("Got here\n"); 

       result = VCE_wait_for_status(&done_machine_type, &completed_subtask);

      printf("Got there\n"); 

       if (result == SUCCESS) 

          kill_slow_machines(done_machine_type, completed_subtask);


       else  { printf("Problem with %s on machine type %d\n", 
                       (*completed_subtask).name,
                       done_machine_type); 

               exit(4);
             }

     }

   printf("*** Processing Complete ***\n");

   VCE_shutdown();
}







